# Main.pm - The Object Class that provides a Main Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: Main.pm,v 1.60 2006/05/12 22:09:52 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::DB::Main;
use strict;
use Portal::Base;
use Portal::Methods;
use Portal::Accounting::Objects::Ledger;
use Portal::Data::Variables;
use Portal::Accounting::Objects::Account;
use Portal::Accounting::Objects::SubAccount;
use Portal::Accounting::Objects::MonitorObject;
use Date::Calc qw(:all);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.60 $';

=head1 NAME

Main - Object used to build a Main Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::DB::Main;
  my $obj = Portal::Accounting::DB::Main->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

Main is a object used to retreive data from the database.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(dbHandle => , lanObj => )

 Creates a new instance of the Portal::Accounting::DB::Main object.
 See Portal::Base(3) for a listing of required arguments.

 requires:
   dbHandle

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( dbHandle => undef, @_ );

  if ($self->error)
  {
    $self->error();
  }

  # instantiate anything unique to this module
  $self->{app} = (split /::/, ref $self)[1];
  $self->{portalVariables} = Portal::Data::Variables->new(langObj => $self->{langObj});
  if ($self->{portalVariables}->error)
  {
    $self->error($self->{portalVariables}->errorMessage);
  }
  eval { $self->{appConfigObj} = Portal::Data::Config->new(langObj => $self->{langObj}, app => $self->{app}); };
  if ($@)
  {
    $self->error($@);
  }
  if ($self->{appConfigObj}->error)
  {
    $self->error($self->{appConfigObj}->errorMessage);
  }
  $self->{dbHandle} = $args{dbHandle};

  $self->{methods} = Portal::Methods->new(langObj => $self->{langObj});
  if ($self->{methods}->error) { $self->error($self->{methods}->errorMessage); }

  # do validation
  if (!$self->Portal::Accounting::DB::Main::isValid)
  {
    # the error is set in the isValid() method.
  }

  # do anything else you might need to do.

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->error();
    return 0;
  }

  # validate our parameters.
  if (!defined $self->{dbHandle})
  {
    $self->error("Missing dbHandle");
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->postfixError($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

#******** Helper functions ***********

=item bool areWeConfigured(void)

    Returns 1 if we are configured, 0 if not, -1 on error.

=cut
sub areWeConfigured
{
  my $self = shift;
  my $result = 0;  # startout saying we aren't configured.
  my $sth;
  my @info;

  $sth = $self->{dbHandle}->read(sql => "SELECT value FROM config_tb WHERE name = 'never_run'");
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return -1;
  }
  @info = $sth->fetchrow_array;
  if (!defined $info[0])
  {
    $self->error(errorString => "nothing returned!<br>\n");
    return -1;
  }
  else
  {
    $result = !$info[0];
  }

  return $result;
}

=item bool sayWeAreConfigured(void)

    Returns 1 on success, 0 on error.

=cut
sub sayWeAreConfigured
{
  my $self = shift;

  $self->{dbHandle}->write(sql => "UPDATE config_tb SET value = '0' WHERE name = 'never_run'");  # set it false.  Looking ahead to days of MySQL possibility.
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  $self->{dbHandle}->commit;  # make change permanent.

  return 1;
}

# logTempFile
# requires: filename
# optional:
# returns:
# This routine makes an entry in the database that indicates a tempfile that needs to be cleaned up later by the cron job.
sub logTempFile
{
  my $self = shift;
  my %args = ( filename => "", @_ );
  my $filename = $args{filename};
  my $sth;
  my @info;

  if (length $filename == 0)
  {
    $self->error(errorString => "filename = '$filename' is invalid!<br>\n");
    return;
  }
  if (! -e $filename)
  {
    $self->error(errorString => "filename = '$filename' does not exist!<br>\n");
    return;
  }

  $sth = $self->{dbHandle}->write(sql => "INSERT INTO temp_file_tb (timeof, fname) VALUES (now(), ?)", plug => [ $filename ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return;
  }
  $self->{dbHandle}->commit;
}


# getTempFilesToCleanup
# requires: timeLength
# optional: timeValue (hours, minutes, seconds, etc.)
# returns: Array of files that need to be removed.
# summary: This method checks the temp_file_tb to see if there are any records that are older than the specified timeLength and if so,
# it returns the list of them in the array.
sub getTempFilesToCleanup
{
  my $self = shift;
  my %args = ( timeLength => 3, timeValue => "hours", @_ );
  my $timeLength = $args{timeLength};
  my $timeValue = $args{timeValue};
  my @files = ();
  my $sth;
  my @info;

  # validate the timeLength and timeValue values
  if ($timeLength < 1)
  {
    $self->error(errorString => "timeLength = '$timeLength' is invalid!<br>\n");
    return @files;
  }
  if ($timeValue !~ /^(hours|minutes|seconds|days|months|years)$/)
  {
    $self->error(errorString => "timeValue = '$timeValue' is invalid!<br>\n");
    return @files;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT fname FROM temp_file_tb WHERE timeof < CAST(now() - CAST('$timeLength $timeValue' AS INTERVAL) AS TIMESTAMP)");
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return @files;
  }

  while (@info = $sth->fetchrow_array)
  {
    $files[++$#files] = $info[0];
  }

  return @files;
}

# cleanupTempFile
# requires: fileName
# optional:
# returns: 1=OK, 0=Error, -1=Doesn't exist on disk (but delete from db was ok), -2=Doesn't exist in db (but did exist on disk and was deleted),
#         -3=Doesn't exist on disk or in db.
# summary: Deletes the specified file and removes it from the temp_file_tb table.
sub cleanupTempFile
{
  my $self = shift;
  my %args = ( fileName => "", @_ );
  my $file = $args{fileName};
  my $result = 1;

  if (length $file == 0)
  {
    $self->error(errorString => "file must be specified!<br>\n");
    return 0;
  }
  if (! -e $file)
  {
    $result = -1;
  }
  unlink $file;

  my $rc = $self->{dbHandle}->write(sql => "DELETE FROM temp_file_tb WHERE fname = ?", plug => [$file]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  if ($rc == 0)
  {
    if ($result == -1)
    {
      $result = -3;  # signal it doesn't exist on disk or in db.
    }
    else
    {
      $result = -2; # it doesn't exist in the db.
    }
  }
  else
  {
    $self->{dbHandle}->commit;
  }

  return $result;
}

# assignAccountToLedger
# requires: accountObj, ledgerCode
# optional:
# returns: 1=OK, 0=Error, -1=Already exists
# summary: assigns the specified account to the specified ledger.
sub assignAccountToLedger
{
  my $self = shift;
  my %args = ( accountObj => undef, ledgerCode => "", @_ );
  my $accountObj = $args{accountObj};
  my $ledgerCode = $args{ledgerCode};
  my $sth;
  my @info;

  if (!defined $accountObj)
  {
    $self->error(errorString => "accountObj must be defined!<br>\n");
    return 0;
  }

  if (length $ledgerCode == 0)
  {
    $self->error(errorString => "ledgerCode must be defined!<br>\n");
    return 0;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT acct_num, ledger_code FROM local_accounts_tb WHERE acct_num = ? AND ledger_code = ?",
                                    plug => [ $accountObj->number, $ledgerCode ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    return -1;  # it already exists in the database!
  }

  my $rc = $self->{dbHandle}->write(sql => "INSERT INTO local_accounts_tb (acct_num, taxable, ledger_code) VALUES (?, ?, ?)",
                                    plug => [ $accountObj->number, $accountObj->taxable, $ledgerCode ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  $self->{dbHandle}->commit;

  return 1;
}

# unAssignAccountFromLedger
# requires: accountObj, ledgerCode
# optional:
# returns: 1=OK, 0=Error, -1=Doesn't exist anymore
# summary: unassigns the specified account from the specified ledger.
sub unAssignAccountFromLedger
{
  my $self = shift;
  my %args = ( accountObj => undef, ledgerCode => "", @_ );
  my $accountObj = $args{accountObj};
  my $ledgerCode = $args{ledgerCode};
  my $sth;
  my @info;

  if (!defined $accountObj)
  {
    $self->error(errorString => "accountObj must be defined!<br>\n");
    return 0;
  }

  if (length $ledgerCode == 0)
  {
    $self->error(errorString => "ledgerCode must be defined!<br>\n");
    return 0;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT acct_num, ledger_code FROM local_accounts_tb WHERE acct_num = ? AND ledger_code = ?",
                                    plug => [ $accountObj->number, $ledgerCode ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (!defined $info[0])
  {
    return -1;  # it doesn't exist in the database!
  }

  my $rc = $self->{dbHandle}->write(sql => "DELETE FROM local_accounts_tb WHERE acct_num = ? AND ledger_code = ?",
                                    plug => [ $accountObj->number, $ledgerCode ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  $self->{dbHandle}->commit;

  return 1;
}

# checkEquityAccount
# requires: accountNumber, ledgerCode
# optional:
# returns: 1 on successful verification of the equity account, 0 if anything goes wrong or the account is invalid
# summary: determines whether the specified account is a valid, active, assigned equity account in the system and available
#          for use by the given ledger
sub checkEquityAccount
{
  my $self = shift;
  my %args = ( accountNumber => "", ledgerCode => "", @_ );
  my $accountName = "";
  my $accountNumber = $args{accountNumber};
  my $ledgerCode = $args{ledgerCode};
  my $sth;
  my @info;

  if (not defined $accountNumber || $accountNumber eq "")
  {
    $self->error(errorString => sprintf($self->{langObj}->map("invalidArgument"), "accountNumber", $accountNumber));
    return 0;
  }
  if (not defined $ledgerCode || $ledgerCode eq "")
  {
    $self->error(errorString => sprintf($self->{langObj}->map("invalidLedger"), $ledgerCode));
    return 0;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT type, active, name FROM account_tb WHERE number = ?",
                                    plug => [ $accountNumber ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    my $type = $info[0];
    my $active = $info[1];
    $accountName = $info[2];

    if ($active =~ /f/i || $active =~ /0/ || $active =~ /false/i)
    {
      $self->error(errorString => sprintf($self->{langObj}->map("accountNotActive"), $accountName, $accountNumber));
      return 0;
    }
    if ($type != 7)
    {
      $self->error(errorString => sprintf($self->{langObj}->map("accountNotAnEquityAccount"), $accountName, $accountNumber));
      return 0;
    }
  }
  else
  {
    # account does not exist
    $self->error(errorString => sprintf($self->{langObj}->map("accountDoesNotExistInDatabase"), $accountName, $accountNumber));
    return 0;
  }

  # check to make sure the equity account is assigned to the specified ledger
  $sth = $self->{dbHandle}->read(sql => "SELECT COUNT(acct_num) FROM local_accounts_tb WHERE acct_num = ? AND ledger_code = ?",
                                    plug => [  $accountNumber, $ledgerCode ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  @info = $sth->fetchrow_array;
  if (not defined $info[0] || $info[0] ne "1")
  {
    $self->error(errorString => sprintf($self->{langObj}->map("accountNotAssignedToLedger"), $accountName, $accountNumber, $ledgerCode));
    return 0;
  }

  # successful equity account
  return 1;
}

=item MonitorObject getMonitorInfo(id)

    Looks up the specified monitor_tb entry.  If not found returns undef,
    else returns the Portal::Accounting::Objects::MonitorObject object.
    If an error occurs, then we set the error and return undef.

=cut
sub getMonitorInfo
{
  my $self = shift;
  my %args = (id => "", @_);
  my $id = $args{id};
  my $info = undef;
  my $sth;
  my @info;

  if ($id !~ /^(\d+)$/)
  {
    $self->error(errorString => "id = '$id' is invalid!");
    return undef;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT xml FROM monitor_tb WHERE id = $id");
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return undef;
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    $info = Portal::Accounting::Objects::MonitorObject->new(id => $id, xml => $info[0], langObj => $self->{langObj});
    if ($info->error)
    {
      $self->error(errorString => $info->errorMessage);
      return undef;
    }
  }

  return $info;
}

=item @MonitorObject getMonitorEntries(ledgerCode)

    Looks up the monitor_tb entries for the specified ledgerCode or all
    entries if ledgerCode = "".  If nothing found returns undef, else
    returns the array of Portal::Accounting::Objects::MonitorObject
    objects.  If an error occurs, then we set the error and return undef.

=cut
sub getMonitorEntries
{
  my $self = shift;
  my %args = (ledgerCode => "", @_);
  my $ledgerCode = $args{ledgerCode};
  my @entries = ();
  my $sth;
  my @info;

  my $sql = "SELECT id, ledger_code, xml FROM monitor_tb";
  if (length $ledgerCode > 0)
  {
    $sql .= " WHERE ledger_code = '$ledgerCode'";
  }
  $sql .= " ORDER BY ledger_code, id";

  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    my $info = Portal::Accounting::Objects::MonitorObject->new(id => $info[0], xml => $info[2], langObj => $self->{langObj});
    if ($info->error)
    {
      $self->error(errorString => $info->errorMessage);
      return undef;
    }
    push @entries, $info;
  }

  return @entries;
}

=item void updateMonitorEntry(monitorObj, action)

    This routine creates or updates an entry in the monitor_tb table that
    represents a lists of accounts to monitor for the specified ledger
    that should email an alert to the specified users/email addresses when
    the specified conditions are met.  If action = "delete" then we delete
    the specified entry.

=cut
sub updateMonitorEntry
{
  my $self = shift;
  my %args = ( monitorObj => undef, action => "", @_ );
  my $monitorObj = $args{monitorObj};
  my $action = $args{action};
  my $rc;
  my @info;

  if (!defined $monitorObj)
  {
    $self->error(errorString => "monitorObj = must be defined!<br>\n");
    return;
  }
  if (!$monitorObj->isValid())
  {
    $self->error(errorString => $monitorObj->errorMessage);
    return;
  }
  if ($action !~ /^(|delete)$/)
  {
    $self->error(errorString => "action = '$action' is invalid!<br>\n");
    return;
  }
  if ($action eq "delete" && $monitorObj->{id} eq "-1")
  {
    $self->error(errorString => "You can not delete an entry with id = '-1'!<br>\n");
    return;
  }

  if ($monitorObj->{id} eq "-1")
  { # create the entry
    my $xml = $monitorObj->export();
    if ($monitorObj->error)
    {
      $self->error(errorString => $monitorObj->errorMessage);
      return;
    }
    $rc = $self->{dbHandle}->write(sql => "INSERT INTO monitor_tb (ledger_code, xml) VALUES (?, ?)",
           plug => [ $monitorObj->{ledgerCode}, $xml ]);
  }
  else
  {
    if ($action eq "delete")
    { # delete the entry
      $rc = $self->{dbHandle}->write(sql => "DELETE FROM monitor_tb WHERE id = ?", plug => [ $monitorObj->{id} ]);
    }
    else
    { # update the entry
      my $xml = $monitorObj->export();
      if ($monitorObj->error)
      {
        $self->error(errorString => $monitorObj->errorMessage);
        return;
      }
      $rc = $self->{dbHandle}->write(sql => "UPDATE monitor_tb SET ledger_code = ?, xml = ? WHERE id = ?",
            plug => [ $monitorObj->{ledgerCode}, $xml, $monitorObj->{id} ]);
    }
  }
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return;
  }
  $self->{dbHandle}->commit;
}

=item scalar getAccountBalance(account, date, ledgerCode)

    requires: account - account number, date - date to balance to,
              ledgerCode - ledger code
    returns:  Account balance.
    date can be defined as 'now()', YYYY-/MM-/DD, MM-/DD-/YYYY

=cut
sub getAccountBalance
{
  my $self = shift;
  my %args = ( account => "", date => "", ledgerCode => "", @_);
  my $account = $args{account};
  my $date = $args{date};
  my $ledgerCode = $args{ledgerCode};
  my $sth;
  my @info;
  my $result = "0.00";  #default to a balance of 0.00

  if ($account !~ /^(\d+)$/)
  {
    $self->error(errorString => "account = '$account' is invalid!");
    return $result;
  }
  if ($ledgerCode !~ /^(.+)$/)
  {
    $self->error(errorString => "ledgerCode = '$ledgerCode' is invalid!");
    return $result;
  }
  if ($date !~ /^(now\(\)|\d{4}[-\/]\d{2}[-\/]\d{2}|\d{2}[-\/]\d{2}[-\/]\d{4})$/)
  {
    $self->error(errorString => "date = '$date' is invalid!");
    return $result;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT account_balance_plus_activity($account, '$ledgerCode', '$date', 'balance') AS balance");
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return $result;
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    $result = $info[0];
  }

  return $result;
}

=item scalar getConfigEntry(name)

    requires: name - name of config_tb table entry to lookup
    returns: undef if name not found or value of entry.

=cut
sub getConfigEntry
{
  my $self = shift;
  my %args = ( name => "", @_ );
  my $name = $args{name};
  my $value = undef;
  my $sth;
  my @info;

  if ($name !~ /^(.+)$/)
  {
    $self->error(errorString => "name = '$name' is invalid!<br>\n");
    return $value;
  }

  $sth = $self->{dbHandle}->read(sql => "SELECT value FROM config_tb WHERE name = ?", plug => [ $name ]);
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return $value;
  }
  @info = $sth->fetchrow_array;
  if (defined $info[0])
  {
    $value = $info[0];
  }

  return $value;
}

=item scalar setConfigEntry(name, value)

    requires: name - name of config_tb table entry to lookup
              value - value to assign to name.
    returns: 1 if ok, 0 on error

=cut
sub setConfigEntry
{
  my $self = shift;
  my %args = ( name => "", value => "", @_ );
  my $name = $args{name};
  my $value = $args{value};
  my $sth;
  my @info;

  if ($name !~ /^(.+)$/)
  {
    $self->error(errorString => "name = '$name' is invalid!<br>\n");
    return 0;
  }

  # see if the entry already exists.
  my $result = $self->getConfigEntry(name => $name);
  if (defined $result)
  {
    # update the entry.
    $self->{dbHandle}->write(sql => "UPDATE config_tb SET value = ? WHERE name = ?", plug => [ $value, $name ]);
  }
  else
  {
    # create the entry.
    $self->{dbHandle}->write(sql => "INSERT INTO config_tb (name, value) VALUES (?, ?)", plug => [ $name, $value ]);
  }
  if ($self->{dbHandle}->error)
  {
    $self->error(errorString => $self->{dbHandle}->errorMessage);
    return 0;
  }
  $self->{dbHandle}->commit;  # make changes permanent.

  return 1;
}

=item string getCashAccountSummary( ledgerCode => , date => )

  Optional:  bank => 1|0
    invest => 1|0
      these flags default to 1.  If set to 0 it will turn off that type of account from showing in the list and total.

  format:  date must be ISO standard YYYY-MM-DD
  returns: HTML <table> output in the event of success;
    undef if something fails

  This routine queries the database for the list of cash/bank accounts.
  Then a table is created containing the current balance, last
  reconciled date, and last entry date for each account.

=cut
sub getCashAccountSummary
{
  my $self = shift;
  my %args = ( ledgerCode => "", date => "", bank => 1, invest => 1, @_ );
  my $ledgerCode = $args{ledgerCode};
  my $date = $args{date};
  my $sth;
  my @info;
  my $htmlString = "";
  my @plug;

  my $phrases = $self->{langObj};
  if ($date eq "")
  {
    $date = POSIX::strftime('%Y-%m-%d', localtime);
  }
  if (length $ledgerCode == 0)
  {
    $self->error(errorString => "ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }
  my $sql;
  if ($args{bank} == 1)
  {
    $sql .= <<"ENDOF";
 SELECT
     a.name as name
  , a.code as code
  , a.number AS id
  , a.type AS type
  , SUM(d.d_amount) as balance
  , (SELECT MAX(trans_date) from debitcredit_tb WHERE account_num = a.number AND ledger_code = d.ledger_code) as last
  , (SELECT MAX(date_of) FROM reconcile_tb WHERE ledger_code = d.ledger_code AND account = a.number) AS reconciled
  FROM account_tb AS a, acct_type_tb AS t, local_accounts_tb AS l, debitcredit_tb AS d
  WHERE
    d.ledger_code = ?
    AND l.ledger_code = d.ledger_code
    AND a.number = l.acct_num
    AND a.type = t.num
    AND a.type = 0
    AND d.account_num = a.number
    AND d.trans_date <= ?
  GROUP BY d.ledger_code, a.type, a.name, a.code, a.number
ENDOF
   push (@plug, $ledgerCode, $date);
 }
 if ($args{invest} == 1)
 {
   if ($sql ne '') { $sql .= " UNION \n"; }
   $sql .= <<"ENDOF";
 SELECT
     a.name as name
  , a.code as code
  , a.number AS id
  , a.type AS type
  , SUM(d.d_amount) as balance
  , (SELECT MAX(trans_date) from debitcredit_tb WHERE account_num = a.number AND ledger_code = d.ledger_code) as last
  , (SELECT MAX(date_of) FROM reconcile_tb WHERE ledger_code = d.ledger_code AND account = a.number) AS reconciled
  FROM account_tb AS a, acct_type_tb AS t, local_accounts_tb AS l, debitcredit_tb AS d
  WHERE
    d.ledger_code = ?
    AND l.ledger_code = d.ledger_code
    AND a.number = l.acct_num
    AND a.type = t.num
    AND a.type = 3
    AND d.sub_account_num = -1
    AND d.account_num = a.number
    AND d.trans_date <= ?
  GROUP BY d.ledger_code, a.type, a.name, a.code, a.number
ENDOF
    push (@plug, $ledgerCode, $date);
  }
  $sql .= "  ORDER BY type, name";
#  return $sql;
  $sth = $self->{dbHandle}->read(sql => $sql, plug => \@plug);
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return undef;
  }

  $htmlString .= "<table class=\"CashSummaryTable\" border=\"0\" cellspacing=\"2\" cellpadding=\"2\">\n";
  $htmlString .= "  <thead><tr>\n";
  $htmlString .= "    <td class=\"CashSummaryName\"><b>".$phrases->map("Account")."</b></td>\n";
  $htmlString .= "    <td class=\"CashSummaryBalance\"><b>".$phrases->map("Cash Balance")."</b></td>\n";
  $htmlString .= "    <td class=\"CashSummaryLastEntry\"><b>".$phrases->map("Last")."</b></td>\n";
  $htmlString .= "    <td class=\"CashSummaryReconcile\"><b>".$phrases->map("Reconcile")."</b></td>\n";
  $htmlString .= "  </tr></thead>\n";

  my $total = 0;
  while (my $account = $sth->fetchrow_hashref)
  {
    my @today = Today();
    #default lookup date for register view to one month ago
    my ($nY,$nM,$nD) = Add_Delta_Days(@today, -31);
    my $start = $nY ."-".($nM - 1) ."-01";
    #if last is greater than 31 days ago use last minus one month
    if ($account->{'last'} ne '')
    {
      my ($year,$month,$day) = split("-",$account->{'last'});
      if (Delta_Days($nY,$nM,$nD,$year,$month,$day) < 0)
      {
        $start = $year."-".($month - 1) ."-01";
      }
    }
    my $class = "positive";
    if ($account->{balance} < 0)
    {
      $class = 'negative';
    }
    $total += $account->{balance};
    my $balance = $self->commify(sprintf("%.2f", $account->{balance}));
    $htmlString .= <<"HTML";
  <tr>
    <td class="CashSummaryName">
      <a href="?app=Accounting&state=Register&command=displayRegister&from=$start&to=$today[0]-$today[1]-$today[2]&ledgerCode=$ledgerCode&account=$account->{id}">$account->{name}</a>
    </td>
    <td class="CashSummaryBalance">
      <span class="$class">$balance</span>
    </td>
    <td class="CashSummaryLastEntry">$account->{'last'}</td>
    <td class="CashSummaryReconcile">$account->{reconciled}</td>
  </tr>
HTML
  }
  my $total = $self->commify(sprintf('%15.2f',$total));
  $htmlString .= <<"HTML";
    <tr>
      <td class="CashSummaryName">&nbsp;</td>
      <td class="CashSummaryBalance"><hr>$total</td>
      <td class="CashSummaryReconcile">&nbsp;</td>
      <td class="CashSummaryLastEntry">&nbsp;</td>
    </tr>
  </table>
HTML

  return $htmlString;
}

=item string getDBVersion

 requires:
 optional:
 returns: Database version or undef if error

=cut
sub getDBVersion
{
  my $self = shift;
  my $Vdb;
  my $sql = "SELECT value FROM config_tb WHERE name = 'dbversion'";
  my $sth = $self->dbHandle->read($sql);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return undef; }
  if (my @d = $sth->fetchrow_array) { $Vdb = $d[0]; }
  else
  {
    $sql = "SELECT version FROM version_tb";
    my $sth = $self->dbHandle->read($sql);
    if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return undef; }
    if (my @d = $sth->fetchrow_array) { $Vdb = $d[0]; }
  }
  return $Vdb;
}

sub commify {
  my $self = shift;
  my $text = reverse $_[0];
  $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
  return scalar reverse $text;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Base(3)

=cut
