# DB::Upgrade.pm - The Object Class that provides a DB::Upgrade Object
#$Id: V2_0.pm,v 1.6 2006/05/12 22:09:52 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::DB::V2_0;
use strict;
use Portal::Accounting::DB::Upgrade;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::DB::Upgrade Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.6 $';

# modify to represent the commands this state provides.
my %commands = (
  display => "Upgrade",
  start => "Upgrade",
  backup => "Backup",
  pre => "Collect Data",
  upgrade => "Upgrade",
  post => "Post Upgrade",
  finish => 'Upgrade Finished',
  topFrame => 'Upgrade Database',
);

=head1 NAME

DB::Upgrade

=head1 SYNOPSIS

  use Portal::Accounting::DB::Upgrade;
  my $obj = Portal::Accounting::DB::Upgrade->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }
  $obj->run(command => $command);

=head1 DESCRIPTION

DB::Upgrade allows for migration from one version of the database to the
next and allows the user to interact with the migration in a nicer
environment than a command line provides.

This is run from the Accounting module when it detects that a new
version of the database is available (code wise) and the database is not
at the version needed.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new()

 Creates a new instance of the Portal::Accounting::DB::Upgrade module.
 See Portal::AppState(3) for a listing of required arguments.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( @_ );

  if ($self->error)
  {
    $self->error('');
  }

  # instantiate anything unique to this module
  $self->{upgradePath} .= "/2.0";
  $self->{upgradeFile} = $self->{upgradePath}."/upgrade.psql";
  $self->{reports} = $self->{upgradePath}."/system_reports.psql";
  # do validation
  if (!$self->isValid)
  {
    # the error is set in the isValid() method.
  }

  # do anything else you might need to do.
  $self->{commands} = \%commands;

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub c_pre
{
  my $self = shift;
  my $doc = $self->prepDoc;
  $doc->print("No data needed");
  $self->dbHandle->write("INSERT INTO config_tb (name, value) VALUES ('nbsp', '&nbsp;')");
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return $doc; }
  return $self->SUPER::c_pre(@_, doc => $doc); #could send command => 'pre' if needed to come back here for more processing
}

sub c_post
{
  use Portal::Accounting::Objects::JournalEntry;
  use Portal::Accounting::Objects::OpenBooks;
  use Portal::Accounting::Objects::CloseOut;
  my $self = shift;
  my $doc = $self->prepDoc;
  my @defaultEquityAccount = ( 3000 );  #this could be a problem for anyone who deleted this account

  # Convert beginning_balance_entry_tb table entries into Journal Entry transactions
  my $sql = <<"ENDOF";
  SELECT
    DISTINCT(beginning_balance_entry_tb.ledger_code),
    (
      MIN(known_years_tb.year) || '-' ||
      (
        SELECT start1 from ledger_tb WHERE code = beginning_balance_entry_tb.ledger_code
      )
    )::date - 1 AS start
  FROM beginning_balance_entry_tb, known_years_tb
  WHERE beginning_balance_entry_tb.ledger_code = known_years_tb.ledger_code
  GROUP BY beginning_balance_entry_tb.ledger_code
ENDOF
  my $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error())
  {
    $self->error("<div>Error obtaining distinct list of used ledger codes from beginning_balance_tb table!</div>" . $self->{dbHandle}->errorMessage);
    $doc->print($self->errorMessage);
    return $doc;
  }
  while (my @info = $sth->fetchrow_array)
  {
    # Create journal entries using the books start date less one day
    my $date = $info[1];
    my $transObj = Portal::Accounting::Objects::JournalEntry->new( dbHandle => $self->{dbHandle}, langObj => $self->{langObj}, ledgerCode => $info[0] );
    if ($transObj->error) { $doc->print($transObj->errorMessage); next; }
    $transObj->set('date', $date );
    if ($transObj->error) { $self->error($transObj->errorMessage); next; }
    my $sql2 = "SELECT account_num, sub_account_num, sub_e_code, amount, dorc FROM beginning_balance_entry_tb WHERE ledger_code = ?";
    my $sth2 = $self->{dbHandle}->read(sql => $sql2, plug => [ $info[0] ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("<div>Error obtaining list of entries from beginning_balance_entry_tb for ledger_code = '" . $info[0] . "'</div>" . $self->{dbHandle}->errorMessage);
      next;
    }

    my $debitSum = 0.00;
    my $creditSum = 0.00;
    my $problemString = "ledgerCode = '" . $info[0] . "'<br>";
    #initialize row 0 for the equity account to be used.
    $transObj->addRow(
      account => $defaultEquityAccount[0],
      dorc => 1,
      amount => 0,
    );
    while (my @info2 = $sth2->fetchrow_array)
    {
      $problemString .= "fiscalPriorDate = '" . $date . "', subEntity = '" . $info2[2] . "', accountNum = '" . $info2[0] . "', subAccountNum = '" . $info2[1] . "', dorc = '" . $info2[4] . "', amount = '" . $info2[3] . "'";
      # build up the JournalEntryObject
      $transObj->addRow(subEntity => "$info2[2]", account => $info2[0], subAccount => $info2[1], dorc => $info2[4], amount => $info2[3]);
      if ($transObj->error)
      {
        $self->error("<div>Error adding row to JournalEntry: $problemString </div>" . $transObj->errorMessage);
        next;
      }

      if ($info2[4] =~ /^(true|1)$/)
      {
        $debitSum += $info2[3];
      }
      else
      {
        $creditSum += $info2[3];
      }

      $problemString .= ", creditSum = '" . $creditSum . "', debitSum = '" . $debitSum . "'.<br>\n";
    }
    if ($self->error ) { next; }
    # add equity account amount to offset beginning_balance_entry_tb amounts
    my $sum = 0.00;
    my $dorc = 0;
    $sum = sprintf("%.2f", $debitSum - $creditSum);
    my $amount = sprintf("%.2f", abs($sum));
    if ($sum < 0.00)
    {
      $dorc = 1;
    }
    $problemString .= "equity sum = '" . $sum . "', equity dorc = '" . $dorc . "', equity amount = '" . $amount . "'.<br>\n";
    # update row 0 with total
    $transObj->{rowObjects}[0]->set('dorc', $dorc);
    $transObj->{rowObjects}[0]->set('amount', $amount);
    if ($transObj->{rowObjects}[0]->error)
    {
      $self->error("<div>Error adding row to JournalEntry: <br> $problemString</div>" . $transObj->{rowObjects}[0]->errorMessage);
      next;
    }

    my $refNum = $transObj->insert(system => 1);
    if ($transObj->error)
    {
      $self->error("<div>Error inserting new transaction!  $problemString</div>" . $transObj->errorMessage);
      next;
    }

    # Delete all beginning_balance_entry_tb entries for the current ledger
    my $sth3 = $self->{dbHandle}->write(sql => "DELETE FROM beginning_balance_entry_tb WHERE ledger_code = ?", plug => [ $info[0] ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("<div>Error deleting entries from beginning_balance_entry_tb table for ledger_code = '" . $info[0] . "'!" . $self->{dbHandle}->errorMessage);
      next;
    }

    # Create new beginning_balance_entry_tb entry using reference number from journal entry and current ledger code
    $sth3 = $self->{dbHandle}->write(sql => "INSERT INTO beginning_balance_entry_tb (ref_num, ledger_code) VALUES (?, ?)", plug => [ $transObj->get('id'), $info[0] ]);
    if ($self->{dbHandle}->error())
    {
      $self->error("<div>Error inserting updated entry into beginning_balance_entry_tb table for ref_num = '" . $transObj->get('id') . "', ledger_code = '" . $info[0] . "'!" . $self->{dbHandle}->errorMessage);
      next;
    }
  }

  #fix account codes to not have spaces
  my $codesql = "SELECT name, number AS id, REPLACE(code, ' ', '') AS code FROM account_tb WHERE code LIKE '% %' ORDER BY code";
  my $sth = $self->{dbHandle}->read(sql => $codesql);
  if ($self->{dbHandle}->error())
  {
    $self->error($self->{dbHandle}->errorMessage);
    $doc->print($self->errorMessage);
    return $doc;
  }
  while (my $info = $sth->fetchrow_hashref)
  {
    my $read = $self->{dbHandle}->read(sql => "SELECT number AS id FROM account_tb WHERE code = '$info->{code}'");
    if ($self->{dbHandle}->error())
    {
      $self->error($self->{dbHandle}->errorMessage);
      $doc->print($self->errorMessage);
      return $doc;
    }
    my $newsql = "UPDATE account_tb SET code = REPLACE(code, ' ', '') WHERE number = $info->{id};\n";
    if (my $foo = $read->fetchrow_arrayref)
    {
      $newsql = "UPDATE account_tb SET code = REPLACE(code, ' ', '')||number WHERE number = $info->{id};\n";
    }
    my $sth2 = $self->{dbHandle}->write(sql => $newsql);
    if ($self->{dbHandle}->error())
    {
      $self->error($self->{dbHandle}->errorMessage);
      $doc->print($self->errorMessage);
      return $doc;
    }
  }

  #fix subaccount codes to not have spaces
  my $sth = $self->{dbHandle}->read(sql => "SELECT sub_acc_name AS name, sub_acc_number AS id, REPLACE(code, ' ', '') AS newcode, code FROM sub_acc_tb WHERE code LIKE '% %'");
  if ($self->{dbHandle}->error())
  {
    $self->error($self->{dbHandle}->errorMessage);
    $doc->print($self->errorMessage);
    return $doc;
  }
  while (my $info = $sth->fetchrow_hashref)
  {
    my $read = $self->{dbHandle}->read(sql => "SELECT sub_acc_number AS id FROM sub_acc_tb WHERE code = '$info->{newcode}'");
    if ($self->{dbHandle}->error())
    {
      $self->error($self->{dbHandle}->errorMessage);
      $doc->print($self->errorMessage);
      return $doc;
    }
    my $newsql = "UPDATE sub_acc_tb SET code = REPLACE(code, ' ', '') WHERE sub_acc_number = $info->{id};\n";
    if (my $foo = $read->fetchrow_arrayref)
    {
      $newsql = "UPDATE sub_acc_tb SET code = REPLACE(code, ' ', '')||sub_acc_number WHERE sub_acc_number = $info->{id};\n";
    }
    my $sth2 = $self->{dbHandle}->write(sql => $newsql);
    if ($self->{dbHandle}->error())
    {
      $self->error($self->{dbHandle}->errorMessage);
      $doc->print($self->errorMessage);
      return $doc;
    }
  }

  use Portal::Accounting::Payroll::Objects::Employee;
  my @objects;
  my %deletes;
  #select employee subs and create new subs for each
  my $subs = <<"ENDOF";
  SELECT
    DISTINCT(d.sub_account_num) AS id,
    s.sub_acc_name AS name,
    s.code,
    d.ledger_code
  FROM debitcredit_tb AS d, sub_acc_tb AS s
  WHERE
    s.sub_acc_number = d.sub_account_num
    AND s.type = 120

  UNION
  SELECT
    sub_acc_number AS id,
    s.sub_acc_name AS name,
    s.code,
    NULL
  FROM sub_acc_tb AS s
  WHERE s.type = 120

  ORDER BY id, ledger_code
ENDOF

  #select all emps used and all emps orderd by id, ledger_code DESC
  #mark each empl with flag of 0
  #as each empl is found to be used in a ledger change flag to 1
  #1s will be deleted
  #0s will be handled specially
  my $sth = $self->dbHandle->read($subs);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return $doc;}
  while (my $info = $sth->fetchrow_hashref)
  {
    if ($info->{ledger_code} ne '')
    {
      my $o = Portal::Accounting::Payroll::Objects::Employee->new(langObj => $self->langObj, dbHandle => $self->dbHandle, ledgerCode => $info->{ledger_code});
      $o->set('name', $info->{name});
      $o->set('fullname', $info->{name});
      $o->set('code', "$info->{code}.$info->{ledger_code}");
      $o->insert();
      if ($o->error) { $self->error($o->errorMessage); return $doc; }
      $o->{oldId} = $info->{id};
      push(@objects, $o);
    }
    $deletes{$info->{id}} = 1;  #flag to delete this record later
  }
  #update debitcredit_tb to use new subs
  foreach my $o (@objects)
  {
    my $update = "UPDATE debitcredit_tb SET sub_account_num = ? WHERE ledger_code = ? AND sub_account_num = ?";
    my $write = $self->dbHandle->write(sql => $update, plug => [$o->get('id'), $o->get('ledgerCode'), $o->{oldId} ]);
    if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return $doc;}
  }
  #delete old subs from sub_acc_tb
  foreach my $d (keys %deletes)
  {
    my $sql = "DELETE FROM sub_acc_tb WHERE sub_acc_number = ?";
    my $delete = $self->dbHandle->write(sql => $sql, plug => [$d]);
    if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return $doc;}
  }

  # indicate to the user that Retained Earnings account was used for all initial balance entries instead of
  # what might have originally been there and that any closeouts were reopened and will need to be closed out again
  if ($self->error) { $doc->print($self->errorMessage); return $doc; }
  else
  {
    $doc->print(sprintf($self->{langObj}->map("initialBalanceUpgrade"), $defaultEquityAccount[0] )  );
  }
  return $self->SUPER::c_post(@_, doc => $doc);
}

=back

=cut
1;
__END__

=head1 NOTE

  See updateTemplate as a template for future DB version updates.

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::AppState(3)

=cut
