# Ledger.pm - The Object Class that provides a Ledger Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: Ledger.pm,v 1.51.4.2 2006/08/08 17:06:08 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Objects::Ledger;
use strict;
use Portal::Accounting::Objects::Base;
use Date::Calc qw(Add_Delta_Days Today);
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Objects::Base Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.51.4.2 $';

=head1 NAME

Ledger - Object used to build a Ledger Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Objects::Ledger;
  my $ledgerObj = Portal::Accounting::Objects::Ledger->new()
  $ledgerObj->populate( . . .)
  if ($ledgerObj->error())
  {
    die $ledgerObj->errorMessage();
  }

=head1 DESCRIPTION

Ledger is a Ledger class.  This represents an entry in the
ledger_tb table in the database.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(langObj => )

    Creates a new instance of Portal::Accounting::Objects::Ledger.
    See Portal::Base(3) for a listing of required arguments.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = ( windowName => "", @_ );

  if ($self->error)
  {
    $self->error();
    return $self;
  }
  my $l = $self->{langObj};

  # instantiate anything unique to this module
  $self->{periods} = {
    names => [qw(annual semiannual quarterly monthly semimonthly biweekly weekly daily)],
    values => [qw(annual semiannual quarterly monthly semimonthly biweekly weekly daily)],
  };
  my @date = Today;
  my $cash = $l->map("Cash");
  my $accrual = $l->map("Accrual");
  my $update = $l->map("Update Month Starts")."-->";
  my $StartDate = $l->map("Start Date");
  $self->{data}->{main} = {
    "command" => { -Type => "hidden",  -Value => "save" },
    "app" => { -Type => "hidden", -Value => "Accounting" },
    "state" => { -Type => "hidden", -Value => "Main" },
    "windowName" => { -Type => "hidden", -Value => $args{windowName}, -DataType => 'string', },
    "new" => { -Type => "hidden", -DataType => 'int', -Value => "1" },
    "color" => { -Label => $l->map("Color"), -Type => "color-picker", -Value => "#3388EE",
      -DataType => 'string',
    },
    "name" => { -Label => $l->map("Ledger Name"), -Type => "text", size => 15,
      -DataType => 'string',
    },
    "code" => { -Label => $l->map("Ledger Code"), -Type => "text", size => 15,
      -DataType => 'string',
    },
    "fiscalYear" => { -Label => $l->map("Fiscal Year"), -Type => "text", size => 6,
      -DataType => 'integer', onchange => 'javascript: this.form.year.value = this.value;'
    },
    "workingyear" => { -Label => $l->map("Current")." ".$l->map("Fiscal Year"), -Type => "hidden",
    },
    "offset1" => { -Type => "hidden", -Value => "0" },
    "year" => { -Label => $l->map("Year"), -Type => "text", size => 4,
      -DataType => 'integer',
    },
    "month" => { -Label => $l->map("Month"), -Type => "text", size => 2,
      -DataType => 'integer', -Value => '01',
    },
    "day" => { -Label => $l->map("Day"), -Type => "text", size => 2,
      -DataType => 'integer', -Value => '01',
    },
    "businessType" => { -Label => $l->map("Business Type"), -Type => "select",
      -Options => $self->{formObj}->createSelectOptions($self->getBusinessTypeNames),
      -DataType => "int"
    },
    "basis" => { -Type => "radio", -DataType => 'string', -Label => $l->map("Basis"), -Options => { names => [ $cash, $accrual ], values => [ qw(cash accrual) ] }, -Value => 'cash'
    },
    "start2" => { -Label => "2", -Type => "text", size => 6, -Value => "02-01",
      -DataType => 'string',
    },
    "start3" => { -Label => "3", -Type => "text", size => 6, -Value => "03-01",
      -DataType => 'string',
    },
    "start4" => { -Label => "4", -Type => "text", size => 6, -Value => "04-01",
      -DataType => 'string',
    },
    "start5" => { -Label => "5", -Type => "text", size => 6, -Value => "05-01",
      -DataType => 'string',
    },
    "start6" => { -Label => "6", -Type => "text", size => 6, -Value => "06-01",
      -DataType => 'string',
    },
    "start7" => { -Label => "7", -Type => "text", size => 6, -Value => "07-01",
      -DataType => 'string',
    },
    "start8" => { -Label => "8", -Type => "text", size => 6, -Value => "08-01",
      -DataType => 'string',
    },
    "start9" => { -Label => "9", -Type => "text", size => 6, -Value => "09-01",
      -DataType => 'string',
    },
    "start10" => { -Label => "10", -Type => "text", size => 6, -Value => "10-01",
      -DataType => 'string',
    },
    "start11" => { -Label => "11", -Type => "text", size => 6, -Value => "11-01",
      -DataType => 'string',
    },
    "start12" => { -Label => "12", -Type => "text", size => 6, -Value => "12-01",
      -DataType => 'string',
    },
    "updateSI" => { -Type => "button", -Value => $update, onclick => "updateStartInfo(this.form);" },
    "fedid" => { -Label => $l->map("Federal Tax ID"), -Type => "text", size => 15,
      -DataType => 'string',
    },
    "stateid" => { -Label => $l->map("State Tax ID"), -Type => "text", size => 15,
      -DataType => 'string',
    },
    "payroll_period" => { -Type => 'select', -DataType => 'string', -Options => $self->{periods}, -Label => "Default Payroll Period"},
  };
  $self->{css}->{main} = <<"ENDOF";
table.editLedger { border: none; }
td.months { text-align: right; padding-left: 4px; }
th.months { text-align: right; padding-left: 4px; }
td.starts {  }
th.starts {  }
.padded { padding: .5em; }
ENDOF
  $self->{template}->{main} = <<"ENDOF";
<form #FORMARGS#>
  #FORMERRORSTRING#
  #FIELD=app#
  #FIELD=state#
  #FIELD=command#
  #FIELD=new#
  <input type="hidden" name="ledgerOpen" value="0">
  #FIELD=windowName#
  <div class="primary">
    <div style="float: right;">
      <div>
        #LABEL=fiscalYear# #FI=fiscalYear#
        Begins on<br>#LABEL=year# #FI=year# #LABEL=month# #FI=month# #LABEL=day# #FI=day#
        <div style="margin: .3em;">#FIELD=updateSI#</div>
      </div>
      <div>
          <div style="font-weight: bold; ">
            <th class="months">#LANG=Month#</th>:
            <th class="starts">#LANG=Start Date#</th>
          </div>
          <div>
            <span class="months">#LABEL=start2#</span>:
            <span class="starts">#FIELD=start2#</span>
          </div>
          <div>
            <span class="months">#LABEL=start3#</span>:
            <span class="starts">#FIELD=start3#</span>
          </div>
          <div>
            <span class="months">#LABEL=start4#</span>:
            <span class="starts">#FIELD=start4#</span>
          </div>
          <div>
            <span class="months">#LABEL=start5#</span>:
            <span class="starts">#FIELD=start5#</span>
          </div>
          <div>
            <span class="months">#LABEL=start6#</span>:
            <span class="starts">#FIELD=start6#</span>
          </div>
          <div>
            <span class="months">#LABEL=start7#</span>:
            <span class="starts">#FIELD=start7#</span>
          </div>
          <div>
            <span class="months">#LABEL=start8#</span>:
            <span class="starts">#FIELD=start8#</span>
          </div>
          <div>
            <span class="months">#LABEL=start9#</span>:
            <span class="starts">#FIELD=start9#</span>
          </div>
          <div>
            <span class="months">#LABEL=start10#</span>:
            <span class="starts">#FIELD=start10#</span>
          </div>
          <div>
            <span class="months">#LABEL=start11#</span>:
            <span class="starts">#FIELD=start11#</span>
          </div>
          <div>
            <span class="months">#LABEL=start12#</span>:
            <span class="starts">#FIELD=start12#</span>
          </div>
      </div>
    </div>
    <div style="float: left;">
      <div class="padded"><b>#LABEL=name#</b>&nbsp;#FI=name#</div>
      <div class="padded"><b>#LABEL=code#</b>&nbsp;#FI=code#</div>
      <div class="padded"><b>#LABEL=color#</b>&nbsp;#FI=color#</div>
      <div class="padded"><b>#LABEL=businessType#</b> #FI=businessType#</div>
      <div class="padded"><b>#LABEL=basis#</b><br> #FI=basis#</div>
      <div class="padded"><b>#LABEL=fedid#</b> #FI=fedid# <font color="green">%</font></div>
      <div class="padded"><b>#LABEL=stateid#</b> #FI=stateid# <font color="green">%</font></div>
      <div class="padded">#LABEL=payroll_period##FIELD=payroll_period#</div>
    </div>
    <div style="clear: both;">
      <input style="clear: both;" type="submit" name="Submit" value="Submit">
    </div>
  </div>
</form>
<p><font color="green">%</font> #LANG=Optional#
ENDOF
  $self->{javascript}->{main} = <<"ENDOF";
ENDOF
  $self->{profile}->{main} = {
    required => [ qw( name code color basis businessType fiscalYear year month day start2 start3 start4 start5 start6 start7 start8 start9 start10 start11 start12 ) ],
    optional => [ qw(fedid stateid payroll_period) ],
    constraints => {
      name => qr/^[\d\w][\s\d\w]*$/,
      code => qr/^[\d\w]+$/,
      color => qr/^\#[A-Fa-f0-9]{6}$/,
      businessType => qr/^\d$/,
      fiscalYear => qr/^\d{4}$/,
      month => qr/^\d{1,2}$/,
      day => qr/^\d{1,2}$/,
      year => qr/^\d{4}$/,
      start2 => qr/^\d{1,2}-\d{1,2}$/,
      start3 => qr/^\d{1,2}-\d{1,2}$/,
      start4 => qr/^\d{1,2}-\d{1,2}$/,
      start5 => qr/^\d{1,2}-\d{1,2}$/,
      start6 => qr/^\d{1,2}-\d{1,2}$/,
      start7 => qr/^\d{1,2}-\d{1,2}$/,
      start8 => qr/^\d{1,2}-\d{1,2}$/,
      start9 => qr/^\d{1,2}-\d{1,2}$/,
      start10 => qr/^\d{1,2}-\d{1,2}$/,
      start11 => qr/^\d{1,2}-\d{1,2}$/,
      start12 => qr/^\d{1,2}-\d{1,2}$/,
    },
  };

  $self->{table} = 'ledger_tb';
  $self->{id} = 'code';
  $self->{INSERTorder}[0] = [ qw( name code color cash_or_accrual business_type fy_mo fy_day start2 start3 start4 start5 start6 start7 start8 start9 start10 start11 start12 fedid stateid payroll_period) ] ;  #insert order for db column names
  $self->{newOnly}[0] = { code => "", day => "", month => '', year => '', fiscal_year => '', start2 => '', start3 => '', start4 => '', start5 => '', start6 => '', start7 => '', start8 => '', start9 => '', start10 => '', start11 => '', start12 => '' };
  $self->{db2data}[0] = {
   business_type => 'businessType',
   cash_or_accrual => 'basis',
   fy_day => 'day',
   fy_mo => 'month',
   fiscal_year => 'fiscalYear',
   };  #map database columns to ->data hash, not specified assumes same name
  $self->{data2db}[0] = {reverse %{$self->{db2data}[0]}};
  $self->{SELECT}[0] = "SELECT ##COLUMNS##, 0 AS new, offset1";
  $self->{SELECT}[0] .= "  , working_year(code) AS workingyear";
  $self->{SELECT}[0] .= "  , (SELECT MIN(year) FROM known_years_tb WHERE ledger_code = code) AS year";
  $self->{SELECT}[0] .= "  , (SELECT MIN(year) FROM known_years_tb WHERE ledger_code = code) + offset1 AS fiscal_year";
  $self->{SELECT}[0] .= " FROM $self->{table} WHERE $self->{id} = '#FIELD=$self->{id}#'";
  $self->{SELECT}[1] .= " SELECT value AS default_payroll_period FROM config_tb WHERE name = 'default_payroll_period'";
  $self->{INSERT}[0] = "INSERT INTO $self->{table} (##COLUMNS##) VALUES (##VALUES##)";
  $self->{DELETE}[0] = "DELETE FROM $self->{table} WHERE code = '#FIELD=$self->{id}#'";
  #what should we output as the value of the thing just deleted?
  $self->{postDELETE}[0] = "SELECT name||', '||code FROM $self->{table} WHERE code = '#FIELD=$self->{id}#' ";
  $self->{LIST}[0] = "SELECT name||' ('||code||')' AS name, code FROM $self->{table} ORDER BY name";

  # do anything else you might need to do.
  $self->{offset} = ();
  return $self;
}

=item boolean populate( . . . )

    new       - 1 for yes, 0 for no
    name      - text
    code      - text
    color     - text
    basis - string (cash, accrual)
    businessType  - integer pertaining to a business type (see business_type_tb)
    fiscalYear -  the first fiscalYear
    year      - the actual year for the first fiscal year
    month     - the first month of the fiscal year
    day       - the first day of the first month
    start2 - start12  - begginning of each month in "MMDD" format

=cut
sub populate
{
  my $self = shift;
  my %input = (fiscalYear => '', year => '', month => '', day => '', workingYear => '', @_);
  $self->{knownYears}->{$input{fiscalYear}} = "$input{year}-$input{month}-$input{day}";
  $self->{offset}[1] = $input{year} - $input{fiscalYear};
  if ($input{workingYear} ne '')
  {
    $self->{workingYear} = $input{workingYear};
  }
  else
  {
    $self->{workingYear} = $input{fiscalYear};
  }

  my $foo = $self->SUPER::populate(@_);
  return $foo;
}

=item bool isNew(void)

  This routine checks the new flag to see if the ledger already exists in the database.
  1 for new, 0 for already exists

=cut
sub isNew
{
  my $self = shift;
  return $self->get('new');
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->error();
    return 0;
  }

  # validate our parameters.

  if ($self->get('offset1') !~ m/^(-?\d+)$/)
  {
    $self->invalid('offset1', $self->get('offset1'));
  }

  my $years = 0;
  for (my $index = 2; $index <= 12; $index++)
  {
    my $d = $self->get("start$index");
    if ($d =~ s/^(\d{2})(\d{2})$/$1-$2/ || $d =~ s/^(\d{1,2})-(\d{1,2})$/sprintf("%02d",$1)."-".sprintf("%02d",$2)/e)
    {
      $self->set("start$index", $d);
      $d =~ s/-//g;
      my $e = $self->start($index - 1);
      $e =~ s/-//g;
      if ($d < $e)
      {
        if ( ++$years > 1 )
        {
          $self->invalid("start".$index, $self->get("start$index : is out of order"));
          $years = 1;
        }
      }
      elsif ($d == $e)
      {
        $self->invalid("start".$index, $self->get("start$index is the same as start".$index - 1));
      }
    }
    else
    {
      $self->invalid("start".$index, $self->get("start$index"));
    }
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

=item String[] businessTypeName(void)

 Returns the business type name from the proper lanquage
 if it is set in the object or an empty string if not set.
 There is no guarantee that it is set!

=cut

sub businessTypeName
{
  my $self = shift;
  for (my $i; $i < @{$self->{data}->{main}->{businessType}->{-Options}->{values}}; $i++)
  {
    if ($self->{data}->{main}->{businessType}->{-Options}->{values}[$i] == $self->get('businessType'))
    {
      return $self->{data}->{main}->{businessType}->{-Options}->{names}[$i];
    }
  }
  return "";
}

=item string end(month => , fiscalYear|delta => ) or end(<month>)

 This routine returns the ending date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

 delta takes an integer that will be added to the current working
 year and used as the year in the result

=cut
sub end
{
  my $self = shift;
  my $month;
  my $year = $self->workingYear();
  my $strip = 0;
  my %args;
  my $return = "";
  if (scalar @_ == 1)
  {
    $month = @_[0];
    $strip = 1;
  }
  else
  {
    %args = (month => "", fiscalYear => $year, delta => 0, @_);
    if ($args{delta} !~ /^-?\d+$/)
    {
      $self->error($self->{langObj}->map("Invalid") . " delta '". $args{delta} . "'<br>\n");
      return undef;
    }
    $month = $args{month};
    $year = $args{fiscalYear} + $args{delta};
  }
  if ($year !~ m/^(\d)+$/)
  {
     $self->error($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("year") . " '". $year . "'<br>\n");
     return undef;
  }
  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
     $self->error($self->{langObj}->map("Invalid") . " " . $self->{langObj}->map("month") . " '". $month . "'<br>\n");
     return undef;
  }
  my @date = split '-', $self->start(month => ($month % 12) + 1, fiscalYear => $year  );
  if ($self->error) { return undef; }
  if ($month == 12)
  {
    $date[0]++;
  }
  if (scalar @date != 3)
  {
    $self->error("Invalid split of date string: ".$self->start(month => $month % 12 + 1, fiscalYear => $year));
    return undef;
  }
  @date = Add_Delta_Days(@date, -1);
  if ($strip)
  {  return sprintf("%02d",$date[1])."-".sprintf("%02d",$date[2]);  }
  return sprintf("%02d",$date[0])."-".sprintf("%02d",$date[1])."-".sprintf("%02d",$date[2]);
}

=item date[] getQuarterEnds([fiscal year])

 Returns an array of strings representing the ending dates of Quarters
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut
sub getQuarterEnds
{
  my $self = shift;
  my @quarters;
  my $year = 0;
  my @month = qw [ 3 6 9 12 ];
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }

  my $index = -1;
  for my $month (@month)
  {
    if ($year)
    {
      $quarters[++$index] = $self->end(month => $month, fiscalYear => $year);
    }
    else
    {
      $quarters[++$index] = $self->end($month);
    }
  }

  return @quarters;
}

=item date[] getQuarterStarts([fiscal year])

 Returns an array of strings representing the starting dates of Quarters
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut
sub getQuarterStarts
{
  my $self = shift;
  my @quarters = ();
  my $year = 0;
  my @month = qw [ 1 4 7 10 ];
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@month)
  {
    $quarters[++$index] = $self->start(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $quarters[$index];
      $quarters[$index] = "$date[1]-$date[2]";
    }
  }

  return @quarters;
}

=item date[] getMonthEnds([fiscal year])

 Returns an array of strings representing the ending dates of months
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getMonthEnds
{
  my $self = shift;
  my @months = qw [ 1 2 3 4 5 6 7 8 9 10 11 12 ];
  my @return;
  my $year = 0;
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@months)
  {
    $return[++$index] = $self->end(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $return[$index];
      $return[$index] = "$date[1]-$date[2]";
    }
  }

  return @return;
}

=item date[] getMonthStarts([fiscal year])

 Returns an array of strings representing the starting dates of months
 for this ledger.  format will be MM-DD unless a year is given to get YYYY-MM-DD.
 -1 for the year will use the current fiscal year.

=cut

sub getMonthStarts
{
  my $self = shift;
  my @months = qw [ 1 2 3 4 5 6 7 8 9 10 11 12 ];
  my @return;
  my $year = 0;
  if (scalar @_ == 1)
  {
    if (@_[0] == -1)
    {  $year = $self->workingYear;  }
    else
    {  $year = @_[0];  }
  }
  if ($year !~ m/^(\d)+$/)
  { return undef; }

  my $index = -1;
  for my $month (@months)
  {
    $return[++$index] = $self->start(month => $month, fiscalYear => $year);
    if ($year == 0)
    {
      my @date = split "-", $return[$index];
      $return[$index] = "$date[1]-$date[2]";
    }
  }

  return @return;
}

=item int offset(month)

 This routine calculates the offset for each month to change the
 fiscal year value into a calendar year value.

=cut

sub offset
{
  my $self = shift;
  my $month = shift;

  if ($month !~ m/^(\d)+$/ || $month < 1 || $month > 12)
  {
    $self->error("<p>Invalid month sent to offset</p>\n");
    return undef;
  }

  if ($month == 1) { return $self->get('offset1'); }
  if (! defined $self->{offset}[$month])
  {
    my $offset = $self->get('offset1');
    $self->{offset}[1] = $offset;
    if ($self->get("start2") < sprintf("%02d",$self->get("month"))."-".sprintf("%02d",$self->get("day")))
    {
      $offset++;
    }
    $self->{offset}[2] = $offset;
    for (my $index = 3; $index <= 12; $index++)
    {
      #compare this date+offset with the last one+offset
      my $foo = $index - 1;
      if ($self->get("start$index") < $self->get("start" . $foo))
      {
        $offset++;
      }
      $self->{offset}[$index] = $offset;
    }
  }
  return $self->{offset}[$month];
}

=item string start(month => , fiscalYear =>) or start(<month>)

 This routine returns the starting date of the month specified.
 If a year is given, it will be translated to actual (YYYY-MM-DD).
 Otherwise format will be MM-DD.

=cut
sub start
{
  my $self = shift;
  my $month;
  my $year = "";
  my %args;
  my $return = "";
  if (@_ == 1)
  {
    $month = @_[0];
  }
  else
  {
    %args = (month => "", fiscalYear => "", @_);
    $month = $args{month};
    $year = $args{fiscalYear};
  }
  if ($month == 1) {  $return = sprintf("%02d",$self->get("month"))."-".sprintf("%02d",$self->get("day")); }
  else { $return = $self->get("start$month"); }
  if ($return !~ m/^(\d{2})-?(\d{2})$/)
  {
    $self->error("Invalid internal month start: '$return'<br>\n");
    return undef;
  }
  $return = $1."-".$2;
  if ($year eq "")
  {  return $return; }
  elsif ($year =~ m/^\d+$/)
  {  return $year + $self->offset($month) ."-".$return;  }
  $self->error("Invalid year: '".$year."'<br>\n");
  return undef;
}

=item ISO-Date firstDate()

 Returns the first date the books were opened for.

=cut

sub firstDate
{
  my $self = shift;

  # try not to calculate it every time.
  if (!defined $self->{firstDate})
  {
    my @years = keys %{$self->{knownYears}};
    my $start1 = $self->get('month')."-".$self->get('day');

    # we build this up based upon the oldest known year + offset of first month and
    # tack on the first month-day entry.
    my $firstDate = ((sort { $a <=> $b } @years)[0] + $self->offset(1)) . "-" . $start1;

    $self->{firstDate} = $firstDate;
  }

  return $self->{firstDate};
}

=item ISO-Date formatDate(string)

 This routine takes an 8 digit date and returns the it formatted in the ISO standard of YYYY-MM-DD
 if a year is specified.  In either case, we have -'s for seperators.

 If an invalid date is specified, we return "" and set an error.

=cut

sub formatDate
{
  my $self = shift;
  my $date = shift;

  if ($date =~ /^(\d{4})(\d{2})(\d{2})$/)
  {
    $date = "$1-$2-$3";
  }
  elsif ($date =~ /^(\d{2})(\d{2})$/)
  {
    $date = "$1-$2";
  }
  elsif ($date !~ /^((\d{4}-)?\d{2}-\d{2})$/)
  {
    $self->invalid("date", $date);
    $date = "";
    $self->error($self->genErrorString("all"));
  }

  return $date;
}

=item integer workingYear()

  Returns the earliest year on the books that is not closed out.

=cut
sub workingYear
{
  my $self = shift;
  return $self->get('workingyear');
}

=item date lastClose()

  Returns the last closed out date.

=cut
sub lastClose
{
  my $self = shift;
  my $ledgerCode = $self->get('code');
  if ($ledgerCode eq "")
  {
    $self->error("<p>Error! no ledgerCode sent to selectLastClose!</p>\n");
    return undef;
  }
  my $sth = $self->{dbHandle}->write(sql => "SET datestyle TO 'ISO'");
  $sth = $self->{dbHandle}->read(sql => "SELECT MAX(date) FROM closeouts WHERE ledger_code = '$ledgerCode'");
  if ($self->{dbHandle}->error)
  {
    $self->error("Error in getLastClose!<br>\n" . $self->{dbHandle}->errorMessage);
    return undef;
  }
  if (my @info = $sth->fetchrow_array)
  {
    $self->{lastClose} = $info[0];
  }
  return $self->{lastClose};
}

=head2 Hash getBusinessTypeNames()

    This routine selects the business type names from the database and returns them as an hash
    of number = name values where the name is looked up from the Language Object so as to be
    Internationalized.

=cut
sub getBusinessTypeNames
{
  my $self = shift;
  my $index = 0;
  my @info;
  my $sth;
  my %types;

  $sth = $self->{dbHandle}->read(sql => "SELECT name, num FROM business_type_tb ORDER BY num");
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $types{$info[1]} = $self->{langObj}->map($info[0]);
  }

  return \%types;
}

=item integer select

  This routine requires that the code be set first. It will
  load records from the database into this object.

=cut
sub select
{
  my $self = shift;
  my $foo = $self->SUPER::select(@_);
  my $cashOrAccrual = $self->get('basis');
  if ($cashOrAccrual =~ /^f|false|0$/)
  {
    $self->set('basis', "accrual");
  }
  else
  {
    $self->set('basis', "cash");
  }
  return $foo;
}

sub printForm
{
  my $self = shift;
  if (!$self->isNew)
  {
    $self->{data}->{main}->{updateSI}->{-Type} = 'hidden';
    $self->{data}->{main}->{fiscalYear}->{-ReadOnly} = 1;
    $self->{data}->{main}->{year}->{-ReadOnly} = 1;
    $self->{data}->{main}->{month}->{-ReadOnly} = 1;
    $self->{data}->{main}->{day}->{-ReadOnly} = 1;
    $self->{data}->{main}->{year}->{-Label} = '';
    $self->{data}->{main}->{month}->{-Label} = '-';
    $self->{data}->{main}->{day}->{-Label} = '-';
  }
  return $self->SUPER::printForm(@_);
}

=item %years selectKnownYears()

    Returns hash of fiscal years as keys with YYYY-MM-DD as data for each key

=cut
sub selectKnownYears
{
  my $self = shift;
  my $ledgerCode = $self->get('code');
  my $sth;
  my @info;

  if (length $ledgerCode == 0)
  {
    $self->error("ledgerCode = '$ledgerCode' is invalid!<br>\n");
    return undef;
  }

  my $sql = "SELECT year, (SELECT month_end('$ledgerCode',12,year)) FROM known_years_tb WHERE ledger_code = '$ledgerCode' ORDER BY year";
  $sth = $self->{dbHandle}->read(sql => $sql);
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return undef;
  }
  while (@info = $sth->fetchrow_array)
  {
    $self->{knownYears}->{$info[0]} = $info[1];
  }

  return %{$self->{knownYears}};
}

=item @array selectKnownYearsArrays()

    Returns an Array of the quarters for use with the date selection widget

=cut
sub selectKnownYearsArrays
{
  my $self = shift;
  my %years = $self->selectKnownYears;
  #select years
  #for each year calculate the data
    #   start_date, end_date, q1_start, q1_end, q2_start, q2_end, q3_start, q3_end, q4_start, q4_end
  my @result;
  my $count = 0;
  foreach my $year (sort keys %years)
  {
    my @start = $self->getQuarterStarts($year);
    my @end = $self->getQuarterEnds($year);
    $result[$count++] = [ $year, $start[0], $end[3], $start[0], $end[0], $start[1], $end[1], $start[2], $end[2], $start[3], $end[3] ];
  }
  return @result;
}

=item string preOpenBooksDate

  This will return the day before the first known year for the current ledger.

=cut
sub preOpenBooksDate
{
  my $self = shift;
  my %years = $self->selectKnownYears();
  my @years = sort keys %years;
  my $year = $years[0] - 1;
  return $self->end(fiscalYear => $year, month => 12)
}

sub insert
{
  my $self = shift;
  my $sql = "SELECT name FROM ledger_tb WHERE code = '".$self->get('code')."'";
  my $sth = $self->dbHandle->read($sql);
  if (my @info = $sth->fetchrow_array)
  {
#    $self->invalid('code');
    $self->error("<p>".$self->langObj->map('Code')." ".$self->langObj->map('already used').":".$self->get('code')."</p>\n");
    return undef;
  }
  $sql = "SELECT name FROM ledger_tb WHERE name = '".$self->get('name')."'";
  $sth = $self->dbHandle->read($sql);
  if (my @info = $sth->fetchrow_array)
  {
#    $self->invalid('name');
    $self->error("<p>".$self->langObj->map('Name')." ".$self->langObj->map('already used').":".$self->get('name')."</p>\n");
    return undef;
  }
  my $basis = 'false';
  if ($self->get('basis') eq 'cash') { $self->set('basis', 'true'); }
  else { $self->set('basis', 'false'); }
  $self->{INSERT}[0] = "INSERT INTO $self->{table} (";
  $self->{INSERT}[0] .= " start1, offset1, offset2, offset3, offset4, offset5, offset6, offset7, offset8, offset9, offset10, offset11, offset12, ";
  $self->{INSERT}[0] .= "##COLUMNS##) ";
  $self->{INSERT}[0] .= "VALUES ('";
  $self->{INSERT}[0] .=   sprintf("%02d",$self->get('month'))."-".sprintf("%02d",$self->get('day'))."'";
  for (my $i = 1; $i <= 12; $i++)
  {
    $self->{INSERT}[0] .= ", '".$self->offset($i)."'";
  }
  $self->{INSERT}[0] .= ", ##VALUES##)";
  $self->{INSERT}[1] = "INSERT INTO known_years_tb (year, closed, ledger_code) VALUES ('#FIELD=fiscalYear#', false, '#FIELD=code#')";
  return $self->SUPER::insert(@_);
}

sub update
{
  my $self = shift;
  my $sql = "SELECT name FROM ledger_tb WHERE name = '".$self->get('name')."' AND code <> '".$self->get('code')."'";
  my $sth = $self->dbHandle->read($sql);
  if (my @info = $sth->fetchrow_array)
  {
    $self->invalid('name', $self->get('name'), $self->langObj->map('Name')." ".$self->langObj->map('already used'));
    return -1;
  }
  my $basis = 'false';
  if ($self->get('basis') eq 'cash') { $basis = 'true'; }
  $self->set('basis', $basis);
  $self->{UPDATE}[0] = "UPDATE $self->{table} SET ##UPDATE## WHERE code = '#FIELD=$self->{id}#'";
  return $self->SUPER::update(@_);
}

sub lite
{
  my $self = shift;
  my %foo;
  $foo{name} = $self->get('name');
  $foo{code} = $self->get('code');
  $foo{color} = $self->get('color');
  return \%foo;
}

=item int doesSystemBalance()

     requires: ledgerCode, beginDate, endDate
     optional: if beginDate/endDate are not given, the balances will be calculated on the remaining date value.
               If neither date value is given, the balance will be calculated on the entire set of transactions
               for the given ledgerCode.
     returns: 1 if system balances, -1 if debits are greater than credits, -2 if credits are greater than debits,
              or 0 if an error occurred
    summary: This method sums all the debits and credits for the transactions in the debitcredit_tb table for the
             given ledgerCode and dates and compares them to see if they are the same.  It will return different
             error codes based on whether debits are greater than credits or vice versa.

=cut
sub doesSystemBalance
{
  my $self = shift;
  my %args = ( ledgerCode => "", beginDate => "", endDate => "", @_ );
  my $ledgerCode = $args{ledgerCode};
  my $beginDate = $args{beginDate};
  my $endDate = $args{endDate};
  my $sth;
  my %sqlString = (
    debits => "SELECT SUM(d_amount) FROM debitcredit_tb WHERE dorc = true AND ledger_code = ?",
    credits => "SELECT SUM(d_amount) FROM debitcredit_tb WHERE dorc = false AND ledger_code = ?",
  );
  my $sqlWhere = "";
  my @plug = ();
  my @sum;

  # put the valid ledgerCode onto the plug array
  push @plug, $ledgerCode;

  # debits
  if ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/) && (defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/))
  {
    $sqlWhere = "AND trans_date BETWEEN ? AND ?";
    push @plug, $beginDate, $endDate;
  }
  elsif (((not defined $beginDate) || (length $beginDate == 0)) && ((defined $endDate) && ($endDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "AND trans_date <= ?";
    push @plug, $endDate;
  }
  elsif (((not defined $endDate) || (length $endDate == 0)) && ((defined $beginDate) && ($beginDate =~ /(\d{4}-?\d{1,2}-?\d{1,2})|(\d{1,2}-?\d{1,2}-?\d{4})/)))
  {
    $sqlWhere = "AND trans_date >= ?";
    push @plug, $beginDate;
  }
  else
  {
    $self->error(sprintf($self->{langObj}->map("invalidDebitDate"), $beginDate, $endDate));
    return 0;
  }
  my $sql = $sqlString{debits} . $sqlWhere . "\nUNION ALL\n" . $sqlString{credits} . $sqlWhere;
  # retrieve the sum of all debit amounts
  $sth = $self->{dbHandle}->read(sql => $sql, plug => [@plug, @plug] );
  if ($self->{dbHandle}->error)
  {
    $self->error($self->{dbHandle}->errorMessage);
    return 0;
  }
  my $count = 0;
  while (my @info = $sth->fetchrow_array)
  {
    $sum[$count++] = $info[0];
  }
  if ($count != 2)
  {
    $self->error('<div class="error">Only 1 record found with sql<div>'.$sql."</div></div>\n");
  }

  # compare to make sure that debits == credits
  return $sum[0] - $sum[1];
}

sub selectCodesArray
{
  my $self = shift;
  my $sql = "SELECT code FROM ledger_tb ORDER BY code";
  my $db = $self->dbHandle;
  my $r = $db->read(sql => $sql);
  if ($db->error) { $self->error($db->errorMessage); return undef; }
  my @a;
  while (my @d = $r->fetchrow_array)
  { $a[scalar @a] = $d[0]; }
  return @a;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:xiwa@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal::Base(3)

=cut
