# Employee.pm - The Object Class that provides a Employee Object
# Created by James Pattie & Jason A. Pattie, 11/07/2000.
#$Id: EmployeePayroll.pm,v 1.1 2006/05/12 14:35:46 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

package Portal::Accounting::Payroll::Objects::EmployeePayroll;
use strict;
use Portal::Accounting::Payroll::Objects::Employee;
use Business::Payroll::US;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT);

require Exporter;

@ISA = qw(Portal::Accounting::Payroll::Objects::Employee Exporter AutoLoader);
@EXPORT = qw();

$VERSION = '$Revision: 1.1 $';

=head1 NAME

Employee - Object used to build a Employee Object Class.

=head1 SYNOPSIS

  use Portal::Accounting::Payroll::Objects::Employee;
  my $obj = Portal::Accounting::Payroll::Objects::Employee->new;
  if ($obj->error())
  {
    die $obj->errorMessage();
  }

=head1 DESCRIPTION

Employee is a Employee class.

=head1 Exported FUNCTIONS

B<NOTE>: I<bool> = 1(true), 0(false)

=over 4

=item scalar new(ticker, name, number, type, market)

 Creates a new instance of the Portal::Accounting::Objects::Employee module.
 See Portal::Accounting::Objects::Base(3) for a listing of required arguments.

 requires:
  id  - The SubAccount number this Employee is associated with
  ledgerCode, name, code,
 optional: 
  gross, married, active

=cut

sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);
  my %args = (ledgerCode => '', count => '', @_ );

  if ($self->error)
  {
    $self->error();
    return $self;
  }
  
  # instantiate anything unique to this module
  my $count = $args{count};
  $self->{name} = "Portal::Accounting::Payroll::Objects::EmployeePayroll";
  my $l = $self->{langObj};

  # do anything else you might need to do.
  $self->{css}->{payroll} = <<"ENDOF";
.gross { width: 6; }
ENDOF
  $self->{profile}->{payroll} = {
    required => [ qw( name code type fullname) ],
    optional => [ qw( gross married active ) ],
    constraints => {
      id => qr/^\d\d+$/,
      type => qr/^120$/,
      name => qr/^[\d\w][\d\s\w&,-]+$/,
      code => qr/^[\d\w.-]+$/,
      gross => qr/^\d*\.\d{2}$/,
    },
  };
  
  $self->{data}->{payroll}->{id} = { -Type => "hidden"};
  $self->{data}->{payroll}->{name} = { -Label => $l->map("Initials"), -Type => "text", -DataType => "text" };
  $self->{data}->{payroll}->{fullname} = { -Label => $l->map("Name"), -Type => "text", -DataType => "text" };
  $self->{data}->{payroll}->{ledgerCode} = { -Value => $args{ledgerCode}, -Type => "hidden"};
  $self->{data}->{payroll}->{gross} = { -Label => $l->map("Normal Gross Pay"), -Type => "text", -DataType => "float" };

  $self->{headers}->{payroll}  = <<"ENDOF";
  <thead class="secondary">  
    <tr>
      <td>
        Employee
      </td>
      <td>
        Gross Pay
      </td>
    </tr>
  </thead>
ENDOF
  $self->{template}->{payroll} = <<"ENDOF";
    <tr>
      <td>
        <input type="hidden" name="id_$count" value="#VALUE=id#" />
        #VALUE=fullname#
      </td>
      <td>
        <input class="gross" type="text" name="gross_$count" value="#VALUE=gross#" />
      </td>
    </tr>
ENDOF

  $self->{defaultForm} = "payroll";

  return $self;
}

=item bool isValid(void)

 Returns 0 or 1 to indicate if the object is valid.
 The error will be available via errorMessage().

=cut

sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->error();
    return 0;
  }

  # validate our parameters.
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }

  return 1;
}

sub delete
{
  my $self = shift;
}

sub insert
{
  my $self = shift;
}

sub printForm
{
  my $self = shift;
  my %args = (form => 'main', count => '', @_);
  my $count = $args{count};
  if ($args{form} eq 'payroll')
  {
    $self->{template}->{payroll} = <<"ENDOF";
    <tr>
      <td>
        <input type="hidden" name="id_$count" value="#VALUE=id#" />
        #VALUE=fullname#
      </td>
      <td>
        <input class="gross" style="width: 8em; text-align: right;" type="text" name="gross_$count" value="#VALUE=gross#" />
      </td>
    </tr>
ENDOF
  }
  my %form = $self->SUPER::printForm(@_);
  $form{css} .= $self->{css}->{payroll};
  return %form;
}

=item xml updatePayroll(state => , date =>)

  builds xml snippet for payroll xml file
  
=cut
sub updatePayroll
{
  my $self = shift;
  my %args = (state => '', date => '', @_);
  my $state = $args{state};
  my $date = $args{date};
  my $sql = "SELECT id, key, value FROM state_payroll WHERE state = ? AND employee = ?";
  #need to make a person to send to modules with id, name, marital
  my $id = $self->get('id');
  my $marital = 'single';
  if ($self->{main}->{married}->{checked}) { $marital = 'married'; }
  my $xml = '    <person id="'. $self->get('id').'" name="'. $self->get('fullname').'" marital="'.$marital."\" >\n";
  #for now we are hard coding us modules information.  after state stuff is implemented maybe we'll add fed configs too
  my $gross = $self->get('gross');
  my $grossYTD = $self->grossYTD($date);
  my $federalYTD = $self->federalYTD($date);
#$self->error("<div>$grossYTD: $federalYTD</div>");
  if ($self->error) { return ''; }
  $xml .= "      <country name=\"US\" method=\"\" gross=\"$gross\" grossYTD=\"$grossYTD\" federalYTD=\"$federalYTD\"";
  #get list of parameters for US module
  #foreach paramter output name value pairs
#allow="$federalAllow" withHold="$fedwithHold"
  my $sth = $self->dbHandle->read(sql => $sql, plug => ['US', $id]);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); }
  else
  {
    while (my $info = $sth->fetchrow_hashref)
    {
      $xml .= " $info->{key}=\"$info->{value}\"";
    }
  }
  $xml .= ">\n";
  #process state for this employee #!!!move to state object.  $so->xml($employee)
  if ($state ne '')
  {
    $xml .= "        <state name=\"$state\" method=\"\" gross=\"$gross\"";
    my $sth = $self->dbHandle->read(sql => $sql, plug => [$state, $id]);
    if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); }
    else
    {
      while (my $info = $sth->fetchrow_hashref)
      {
        $xml .= " $info->{key}=\"$info->{value}\"";
      }
    }
    $xml .= ">\n        </state>\n";
  }

  $xml .= "      </country>\n";
  
  $xml .= "    </person>\n";
  return $xml;
}

=item float grossYTD(date)

  This routine uses the date given to gather all Salaries paid to the employee represented by this object in the current calendar year.
  
=cut
sub grossYTD
{
  my $self = shift;
  my $to = shift;
  my $from;
  if ($to !~ /^(\d{4})-(\d{1,2})-(\d{1,2})$/) { $self->error("<div>Invalid date '$to' sent to grossYTD</div>"); return undef; }
  else { $from = "$1-01-01"; }
  my $sql = "SELECT sum(d.d_amount) FROM debitcredit_tb AS d, account_tb AS a WHERE d.account_num = a.number AND d.trans_date BETWEEN ? AND ? AND d.sub_account_num = ? AND a.type = 55";
  my $sth = $self->dbHandle->read(sql => $sql, plug => [$from, $to, $self->get('id')]);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); }
  else
  {
    if (my @info = $sth->fetchrow_array)
    {
      if ($info[0] ne '') { return $info[0]; }
    }
  }
  return '0.00';
}

sub federalYTD
{
  my $self = shift;
  my $to = shift;
  my $from;
  if ($to !~ /^(\d{4})-(\d{1,2})-(\d{1,2})$/) { $self->error("<div>Invalid date '$to' sent to federalYTD</div>"); return undef; }
  else { $from = "$1-01-01"; }
  my $sql = "SELECT sum(d.d_amount) FROM debitcredit_tb AS d, account_tb AS a WHERE d.account_num = a.number AND d.trans_date BETWEEN ? AND ? AND d.sub_account_num = ? AND a.type = 24";
  my $sth = $self->dbHandle->read(sql => $sql, plug => [$from, $to, $self->get('id')]);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); }
  else
  {
    if (my @info = $sth->fetchrow_array)
    {
      if ($info[0] ne '') { return $info[0]; }
    }
  }
  return '0.00';
}

sub checkEmployee
{
  my $self = shift;
  my %args = (state => '', @_);
  my $state = $args{state};
  my $sql = "SELECT id, key, value FROM state_payroll WHERE state = ? AND employee = ?";
  my $sth = $self->{dbHandle}->read(sql => $sql, plug => [$state, $self->get('id') ]);
  if ($self->dbHandle->error) { $self->error($self->dbHandle->errorMessage); return 0; }
  if ($sth->fetchrow_hashref) { return 1; }
  return 0;
}

=back

=cut

1;
__END__

=head1 NOTE

 All data fields are accessible by specifying the object
 and pointing to the data member to be modified on the
 left-hand side of the assignment.
 Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin@pcxperience.com)

=head1 SEE ALSO

perl(1), Portal(3), Portal::Accounting(3), Portal::Accounting::Objects::Base(3)

=cut
